// ----------------------------------------------------------------------------
// ...\ImageEdit\Main.pas
//
// Purpose:
//   Main form of application. Image manipulation, of JPEG, PNG and BMP files.
//
// Change history:
//   2019-10-26  Foersom  Created.
//
// License
//   CC-BY 4.0
// ----------------------------------------------------------------------------

unit Main;


interface  // -----------------------------------------------------------------

uses
  PNGImage, JPEG,
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Vcl.ExtCtrls, Vcl.ExtDlgs;

type
  TMainForm = class(TForm)
    ExitButton: TButton;
    Image1: TImage;
    EditFileName: TEdit;
    ButtonSave: TButton;
    ButtonOpen: TButton;
    ButtonFlipHoriz: TButton;
    ButtonFlipVertic: TButton;
    ButtonTurnRight: TButton;
    ButtonTurnLeft: TButton;
    ButtonGrayScale: TButton;
    ButtonCircles: TButton;
    ButtonUndo: TButton;
    OpenDialog1: TOpenDialog;
    SavePictureDialog1: TSavePictureDialog;
    procedure ExitButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ButtonSaveClick(Sender: TObject);
    function LoadImage(FName: TFileName): integer;
    procedure FormDestroy(Sender: TObject);
    procedure ButtonOpenClick(Sender: TObject);
    procedure ButtonFlipHorizClick(Sender: TObject);
    procedure ButtonFlipVerticClick(Sender: TObject);
    procedure ButtonTurnRightClick(Sender: TObject);
    procedure ButtonTurnLeftClick(Sender: TObject);
    procedure ButtonGrayScaleClick(Sender: TObject);
    procedure ButtonCirclesClick(Sender: TObject);
    procedure ButtonUndoClick(Sender: TObject);
  private
    { Private declarations }
    PNG: TPNGImage;  // For load / save of PNG file
    JPEG: TJPEGImage;  // For load / save of JPEG file
    BMPBackup: TBitmap;  // Backup of previous edit step. Used by Undo
    BMPWork: TBitmap;  // For edit of any image and load / save of BMP file
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;


implementation  // ------------------------------------------------------------

{$R *.dfm}

const
  ColorTabMax=10;
  ColorTab: array[0..ColorTabMax-1] of TColor=
      (ClBlack, ClMaroon, ClRed, ClWebDarkOrange, ClYellow,
       ClGreen, ClBlue, ClPurple, ClGray, ClWhite);


function TMainForm.LoadImage(FName: TFileName): integer;
var
  Ext: TFileName;
begin
  result:=0;
  Ext:=UpperCase(ExtractFileExt(FName));

  try
    if (Ext='.JPG') or (Ext='.JPEG') then
    begin
      JPEG.LoadFromFile(FName);  // load to JPEG object
      BMPWork.Assign(JPEG);  // Copy JPEG object to bitmap object
    end
    else if (Ext='.PNG') then
    begin
      PNG.LoadFromFile(FName);  // load to JPEG object
      BMPWork.Assign(PNG);   // Copy PNG object to bitmap object
    end
    else if (Ext='.BMP') then
    begin
      BMPWork.LoadFromFile(FName);  // Load directly to bitmap object
    end
    else
      result:=99999;  // Error, unknown extension
  except
    On E: EInOutError do
      result:=E.ErrorCode;
  end;

  if result=0 then
    Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ButtonOpenClick(Sender: TObject);
var
  FName: TFileName;
  Err: integer;
begin
  FName:=trim(EditFileName.Text);
  if FName<>'' then
    OpenDialog1.InitialDir:=ExtractFilePath(FName)
  else
    OpenDialog1.InitialDir:=GetCurrentDir;
  OpenDialog1.Filter:='Image files|*.png;*.jpg;*.jpeg;*.bmp|All|*.*';
  OpenDialog1.FilterIndex:=1;
  OpenDialog1.Execute;
  FName:=OpenDialog1.FileName;
  if FileExists(FName) then
  begin
    EditFileName.Text:=FName;
    Err:=LoadImage(FName);

    if Err<>0 then
      MessageDlg(format('Error loading image: %s',[FName]),mtError,[mbOK],0);
  end;
end;


procedure TMainForm.ButtonSaveClick(Sender: TObject);
var
  FName: TFileName;
  Ext: TFileName;
  Err, MR: integer;
begin
  Err:=0;
  FName:=trim(EditFileName.Text);

  if FName<>'' then
    SavePictureDialog1.FileName:=FName
  else
    SavePictureDialog1.FileName:=GetCurrentDir;
  SavePictureDialog1.Filter:='Image files|*.png;*.jpg;*.jpeg;*.bmp|All|*.*';
  SavePictureDialog1.FilterIndex:=1;
  SavePictureDialog1.Execute;
  FName:=SavePictureDialog1.FileName;
  EditFileName.Text:=FName;

  MR:=mrOK;
  if FileExists(FName) then
    MR:=MessageDlg('File already exists: '+FName+SLineBreak+SLineBreak+'Overwrite file?',mtWarning,[mbOK,mbCancel],0);

  if mr=mrOK then
  begin
    Ext:=UpperCase(ExtractFileExt(FName));
    try
      if (Ext='.JPG') or (Ext='.JPEG') then
      begin
        JPEG.Assign(BMPWork);  // Copy bitmap image to JPEG object
        JPEG.SaveToFile(FName);
      end
      else if (Ext='.PNG') then
      begin
        PNG.Assign(BMPWork);   // Copy bitmap image to PNG object
        PNG.SaveToFile(FName);
      end
      else if (Ext='.BMP') then
      begin
        BMPWork.SaveToFile(FName);  // Save directly from bitmap object
      end
      else
       Err:=99999;  // Error, unknown extension
    except
      On E: EInOutError do
        Err:=E.ErrorCode;
    end;

    if Err<>0 then
      MessageDlg(format('Error %d saving: %s',[Err,FName]),mtError,[mbOK],0);
  end;
end;


// Draw 20 random circles (ellipses) on the image
procedure TMainForm.ButtonCirclesClick(Sender: TObject);
var
  I: Integer;
  X,Y,R: integer;
  C: TColor;
begin
  BMPBackup.Assign(BMPWork);  // Backup current image
  BMPWork.Canvas.Brush.Style:=bsClear;  // No circle fill
  BMPWork.Canvas.Pen.Width:=4;  // Circle line width
  for I := 1 to 20 do
  begin
    X:=random(BMPWork.Width);
    Y:=random(BMPWork.Height);
    R:=random(80)+20;  // radius 20..99
    C:=ColorTab[I mod ColorTabMax];  // cycle through all colors
    BMPWork.Canvas.Pen.Color:=C;
    BMPWork.Canvas.Ellipse(X-R,Y-R,X+R,Y+R);
  end;
  Image1.Picture.Assign(BMPWork);  // Display image
end;


// Mirror the image
procedure TMainForm.ButtonFlipHorizClick(Sender: TObject);
var
  X,Y,W,H: integer;
begin
  BMPBackup.Assign(BMPWork);  // Backup current image

  H:=BMPBackup.Height;
  W:=BMPBackup.Width;
  for Y := 0 to H-1 do
    for X := 0 to W-1 do
      BMPWork.Canvas.Pixels[W-X-1,Y]:=BMPBackup.Canvas.Pixels[X,Y];

  Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ButtonFlipVerticClick(Sender: TObject);
var
  X,Y,W,H: integer;
begin
  BMPBackup.Assign(BMPWork);  // Backup current image

  H:=BMPBackup.Height;
  W:=BMPBackup.Width;
  for Y := 0 to H-1 do
    for X := 0 to W-1 do
      BMPWork.Canvas.Pixels[X,H-Y-1]:=BMPBackup.Canvas.Pixels[X,Y];

  Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ButtonGrayScaleClick(Sender: TObject);
var
  X,Y,W,H: integer;
  C: TColor;
  R,G,B,Lumi: integer;
begin
  BMPBackup.Assign(BMPWork);  // Backup current image

  H:=BMPBackup.Height;
  W:=BMPBackup.Width;
  for Y := 0 to H-1 do
    for X := 0 to W-1 do
    begin
      C:=BMPBackup.Canvas.Pixels[X,Y];  // bit 23..0 = RGB
      R:=(C shr 16) and $ff;
      G:=(C shr 8) and $ff;
      B:=C and $ff;

      // PAL color luminans = 0.299*R + 0.587*G + 0.114*B
      // Using fractions of 256 for speedy integer calculations
      Lumi:=R*256*299 div 1000 +
            G*256*587 div 1000 +
            B*256*114 div 1000;
      Lumi:=Lumi shr 8;  // div 256 so Luminans in [0..255]
      C:=Lumi shl 16 + Lumi shl 8 + Lumi;  // Set RGB channels to same luminans for grayscale
      BMPWork.Canvas.Pixels[X,Y]:=C;
    end;

  Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ButtonTurnLeftClick(Sender: TObject);
var
  X,Y,W,H: integer;
begin
  BMPBackup.Assign(BMPWork);  // Backup current image

  H:=BMPBackup.Height;
  W:=BMPBackup.Width;
  BMPWork.Height:=W;
  BMPWork.Width:=H;
  for Y := 0 to H-1 do
    for X := 0 to W-1 do
      BMPWork.Canvas.Pixels[Y,W-X-1]:=BMPBackup.Canvas.Pixels[X,Y];

  Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ButtonTurnRightClick(Sender: TObject);
var
  X,Y,W,H: integer;
begin
  BMPBackup.Assign(BMPWork);  // Backup current image

  H:=BMPBackup.Height;
  W:=BMPBackup.Width;
  BMPWork.Height:=W;
  BMPWork.Width:=H;
  for Y := 0 to H-1 do
    for X := 0 to W-1 do
      BMPWork.Canvas.Pixels[H-Y-1,X]:=BMPBackup.Canvas.Pixels[X,Y];

  Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ButtonUndoClick(Sender: TObject);
begin
  BMPWork.Assign(BMPBackup);  // Set current image to previous image
  Image1.Picture.Assign(BMPWork);  // Display image
end;


procedure TMainForm.ExitButtonClick(Sender: TObject);
begin
  Close;
end;


procedure TMainForm.FormCreate(Sender: TObject);
begin
  PNG:=TPNGImage.Create();
  JPEG:=TJPEGImage.Create();
  BMPBackup:=TBitmap.Create;
  BMPBackup.PixelFormat := pf32bit;
  BMPWork:=TBitmap.Create;
  BMPWork.PixelFormat := pf32bit;
  BMPWork.Width:=Image1.Width;
  BMPWork.Height:=Image1.Height;

  Randomize;
end;


procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(BMPWork);
  FreeAndNil(BMPBackup);
  FreeAndNil(JPEG);
  FreeAndNil(PNG);
end;


end.

